
/*
 *     File Name     : CicsEci.cpp
 *
 *     Description   : This sample shows how a mixed mode DLL can be used to
 *                     allow .NET applications to interact with CICS using the
 *                     CICS Transaction Gateway ECI API for C.
 *
 *                     This file contains the implementations of the following
 *                     managed types:
 *                          * EciException class
 *                          * CicsSystem structure
 *                          * CicsEci class
 *
 * Licensed Materials - Property of IBM
 * 
 * (c) Copyright IBM Corporation 2007, 2008
 * All Rights Reserved
 * 
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with
 * IBM Corporation
 *
 *         The following code is sample code created by IBM Corporation.  This
 *         sample code is not part of any standard IBM product and is provided
 *         to you solely for the purpose of assisting you in the development of
 *         your applications.  The code is provided 'AS IS', without warranty
 *         or condition of any kind.  IBM shall not be liable for any damages
 *         arising out of your use of the sample code, even if IBM has been
 *         advised of the possibility of such damages.
 */

#include "stdafx.h"     //Standard include added by Visual Studio
#include <memory.h>     //For the memset function
#include "CicsEci.h"


#ifdef __cplusplus_cli

//Importing the EciWrapper namespace removes the need to prefix every member
//defined in the following code with EciWrapper::.
using namespace EciWrapper;


/*
 *      EciException class
 */
//ReturnCode property
EciReturnCode EciException::ReturnCode::get() {
    return this->rc;
}

//AbendCode property
String^ EciException::AbendCode::get() {
    return this->abCode;
}

//The constructor calls the constructor of the base Exception class to set the
//exception message.
EciException::EciException(short retCode, String^ abendCode) : Exception("The CICS ECI call failed.") {
    this->rc = (EciReturnCode) retCode;
    this->abCode = abendCode;
}


/*
 *      CicsSystem structure
 */
//Name property
String^ CicsSystem::Name::get() {
    return this->sysName;
}

//Description property
String^ CicsSystem::Description::get() {
    return this->sysDesc;
}

//Constructor
CicsSystem::CicsSystem(String^ name, String^ desc) {
    this->sysName = name;
    this->sysDesc = desc;
}


/*
 *      CicsEci class
 */
//CallType property - wraps eci_call_type
EciCallType CicsEci::CallType::get() {
    return (EciCallType) this->eciParms->eci_call_type;
}
void CicsEci::CallType::set(EciCallType callType) {
    this->eciParms->eci_call_type = (cics_sshort_t) callType;
}

//ProgramName property - wraps eci_program_name
String^ CicsEci::ProgramName::get() {
    return CicsEci::NtvCharToMgdString(this->eciParms->eci_program_name, 8);
}
void CicsEci::ProgramName::set(String^ programName) {
    CicsEci::MgdStringToNtvChar(programName, this->eciParms->eci_program_name, 8);
}

//UserID property - wraps eci_userid
String^ CicsEci::UserID::get() {
    return CicsEci::NtvCharToMgdString(this->eciParms->eci_userid, 8);
}
void CicsEci::UserID::set(String^ userID) {
    CicsEci::MgdStringToNtvChar(userID, this->eciParms->eci_userid, 8);
}

//Password property - wraps eci_password
String^ CicsEci::Password::get() {
    return CicsEci::NtvCharToMgdString(this->eciParms->eci_password, 8);
}
void CicsEci::Password::set(String^ password) {
    CicsEci::MgdStringToNtvChar(password, this->eciParms->eci_password, 8);
}

//TransactionID property - wraps eci_transid
String^ CicsEci::TransactionID::get() {
    return CicsEci::NtvCharToMgdString(this->eciParms->eci_transid, 4);
}
void CicsEci::TransactionID::set(String^ transID) {
    CicsEci::MgdStringToNtvChar(transID, this->eciParms->eci_transid, 4);
}

//AbendCode property - wraps eci_abend_code
String^ CicsEci::AbendCode::get() {
    return CicsEci::NtvCharToMgdString(this->eciParms->eci_abend_code, 4);
}

//Commarea property - wraps eci_commarea
array<Byte>^ CicsEci::Commarea::get() {
    return this->commareaArray;
}
void CicsEci::Commarea::set(array<Byte>^ commarea) {
    this->commareaArray = commarea;
}

//Timeout property - wraps eci_timeout
short CicsEci::Timeout::get() {
    return this->eciParms->eci_timeout;
}
void CicsEci::Timeout::set(short timeout) {
    this->eciParms->eci_timeout = timeout;
}

//ExtendMode property - wraps eci_extend_mode
EciExtendMode CicsEci::ExtendMode::get() {
    return (EciExtendMode) this->eciParms->eci_extend_mode;
}
void CicsEci::ExtendMode::set(EciExtendMode extendMode) {
    this->eciParms->eci_extend_mode = (short) extendMode;
}

//Handle property - wraps eci_async_notify
IntPtr CicsEci::Handle::get() {
    return IntPtr(this->eciParms->eci_async_notify.sem_handle);
}
void CicsEci::Handle::set(IntPtr handle) {
    this->eciParms->eci_async_notify.sem_handle = (cics_lhandle_t) handle.ToPointer();
}

//MessageID property - wraps eci_message_id
unsigned short CicsEci::MessageID::get() {
    return this->eciParms->eci_message_id;
}
void CicsEci::MessageID::set(unsigned short messageID) {
    this->eciParms->eci_message_id = messageID;
}

//MessageQualifier property - wraps eci_message_qualifier
short CicsEci::MessageQualifier::get() {
    return this->eciParms->eci_message_qualifier;
}
void CicsEci::MessageQualifier::set(short messageQualifier) {
    this->eciParms->eci_message_qualifier = messageQualifier;
}

//LUWToken property - wraps eci_luw_token
int CicsEci::LUWToken::get() {
    return this->eciParms->eci_luw_token;
}
void CicsEci::LUWToken::set(int luwToken) {
    this->eciParms->eci_luw_token = luwToken;
}

//SysID property - wraps eci_sysid
String^ CicsEci::SysID::get() {
    return CicsEci::NtvCharToMgdString(this->eciParms->eci_sysid, 4);
}
void CicsEci::SysID::set(String^ sysID) {
    CicsEci::MgdStringToNtvChar(sysID, this->eciParms->eci_sysid, 4);
}

//SystemName property - wraps eci_system_name
String^ CicsEci::SystemName::get() {
    return CicsEci::NtvCharToMgdString(this->eciParms->eci_system_name, 8);
}
void CicsEci::SystemName::set(String^ systemName) {
    CicsEci::MgdStringToNtvChar(systemName, this->eciParms->eci_system_name, 8);
}

//Callback property - wraps eci_callback
EciCallback^ CicsEci::Callback::get() {
    return this->callbackDelegate;
}
void CicsEci::Callback::set(EciCallback^ callback) {
    this->callbackDelegate = callback;
}

//TPN property - wraps eci_tpn
String^ CicsEci::TPN::get() {
    return CicsEci::NtvCharToMgdString(this->eciParms->eci_tpn, 4);
}
void CicsEci::TPN::set(String^ tpn) {
    CicsEci::MgdStringToNtvChar(tpn, this->eciParms->eci_tpn, 4);
}

//Constructor
CicsEci::CicsEci() {
    //Create the unmanaged ECI_PARMS structure
    this->eciParms = new ECI_PARMS;
    memset(this->eciParms, 0, sizeof(ECI_PARMS));

    //Set ECI version
    this->eciParms->eci_version = ECI_VERSION_1A;

    //Set callback to null
    this->callbackDelegate = nullptr;
}

//Destructors in C++/CLI compile to a Dispose() method and cause the containing
//type to implement IDisposable. This allows user code to explicitly dispose of
//an object when it is no longer needed without waiting for the garbage
//collector.
CicsEci::~CicsEci() {
    //Delete ECI_PARMS
    delete this->eciParms;

    //Set callback to null
    this->callbackDelegate = nullptr;
}

//The MgdStringToNtvChar utility method copies the contents of a managed String
//object into a given char buffer, up to the size of the buffer.
void CicsEci::MgdStringToNtvChar(String^ src, cics_char_t* dest, int len) {
    cics_char_t*   ntvChars;

    //Zero destination block
    memset(dest, 0, len);

    if (src != nullptr) {
        //Get unmanaged char array
        ntvChars = (cics_char_t*) Marshal::StringToHGlobalAnsi(src).ToPointer();

        //Copy to destination pointer
        if (src->Length > len) {
            memcpy(dest, ntvChars, len);
        } else {
            memcpy(dest, ntvChars, src->Length);
        }

        //Free temp memory
        Marshal::FreeHGlobal((IntPtr) ntvChars);
    }
}

//The NtvCharToMgdString utility method creates a managed String object from a
//given unmanaged char buffer. The length of the text in the buffer is measured
//then a new String object is created and trailing whitespace is removed.
String^ CicsEci::NtvCharToMgdString(cics_char_t* src, int len) {
    String^ mgdString;

    //Measure the string up to a null character or the maximum length len
    int realLen = 0;

    while ((realLen < len) && (src[realLen] != 0)) {
        realLen++;
    }

    //Create managed string and trim off any trailing whitespace
    mgdString = gcnew String(src, 0, realLen);
    return mgdString->TrimEnd(' ');
}

//The ListSystems method calls CICS_EciListSystems with an array large enough
//to hold 20 CICS systems. If the return code is ECI_ERR_MORE_SYSTEMS, the
//array size is increased to match the actual number of systems and the call
//is repeated. All error return codes except ECI_ERR_NO_SYSTEMS cause an
//EciException to be thrown. If there are no systems defined the method will
//continue as normal and an empty array will be returned.
array<CicsSystem>^ CicsEci::ListSystems() {
    array<CicsSystem>^  mgdSysArray;
    CICS_EciSystem_t*   ntvSysArray;
    unsigned short      servCount;
    short               ret;

    //Allocate space for 20 systems
    servCount = 20;
    ntvSysArray = new CICS_EciSystem_t[servCount];
    memset(ntvSysArray, 0, servCount * sizeof(CICS_EciSystem_t));

    //Make the list systems native call
    ret = CICS_EciListSystems(NULL, &servCount, ntvSysArray);

    //Test return code
    if (ret == ECI_ERR_MORE_SYSTEMS) {
        //More than 20 systems, so increase array size and try again
        delete [] ntvSysArray;
        ntvSysArray = new CICS_EciSystem_t[servCount];
        ret = CICS_EciListSystems(NULL, &servCount, ntvSysArray);
    }

    if ((ret != ECI_NO_ERROR) && (ret != ECI_ERR_NO_SYSTEMS)) {
        //An error occurred, throw an exception
        delete [] ntvSysArray;
        throw gcnew EciException(ret, String::Empty);
    }

    //Convert CICS_EciSystem_t array to CicsSystem array
    mgdSysArray = gcnew array<CicsSystem>(servCount);

    for (int i = 0; i < servCount; i++) {
        mgdSysArray[i] = CicsSystem(
                        CicsEci::NtvCharToMgdString(ntvSysArray[i].SystemName, CICS_ECI_SYSTEM_MAX),
                        CicsEci::NtvCharToMgdString(ntvSysArray[i].Description, CICS_ECI_DESCRIPTION_MAX)
                      );
    }

    //Delete native array
    delete [] ntvSysArray;

    //Return managed array
    return mgdSysArray;
}

//The Execute method prepares eci_commarea, eci_commarea_length and
//eci_callback members of the ECI_PARMS structure before making the
//CICS_ExternalCall function call. An additional reference to the commarea
//memory is kept in the form of a pinned pointer, which prevents the garbage
//collector from moving or freeing the memory until the pointer goes out of
//scope at the end of the method. This allows unmanaged code to access the
//data referenced by the pointer safely.
void CicsEci::Execute() {
    short           ret;
    pin_ptr<Byte>   commareaPtr;

    //Pin the managed commarea and assign to ECI_PARMS
    if (this->commareaArray != nullptr) {
        commareaPtr = &this->commareaArray[0];
        this->eciParms->eci_commarea = commareaPtr;
        this->eciParms->eci_commarea_length = this->commareaArray->Length;
    }

    //Set up the callback if needed
    if (this->callbackDelegate != nullptr) {
        this->eciParms->eci_callback = (CICS_EciNotify_t) Marshal::GetFunctionPointerForDelegate(this->callbackDelegate).ToPointer();
    }

    //Make the CICS call
    ret = CICS_ExternalCall(this->eciParms);

    //Check the return code
    if (ret != ECI_NO_ERROR) {
        //Throw an exception
        throw gcnew EciException(ret, this->AbendCode);
    }
}

#endif