
/*
 *     File Name     : CicsEci_2003.cpp
 *
 *     Description   : This sample shows how a mixed mode DLL can be used to
 *                     allow .NET applications to interact with CICS using the
 *                     CICS Transaction Gateway ECI API for C.
 *
 *                     This file contains the implementations of the following
 *                     managed types:
 *                          * EciException class
 *                          * CicsSystem structure
 *                          * CicsEci class
 *
 *                     This file is the Managed C++ equivalent of CicsEci.cpp.
 *
 * Licensed Materials - Property of IBM
 * 
 * (c) Copyright IBM Corporation 2007, 2008
 * All Rights Reserved
 * 
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with
 * IBM Corporation
 *
 *         The following code is sample code created by IBM Corporation.  This
 *         sample code is not part of any standard IBM product and is provided
 *         to you solely for the purpose of assisting you in the development of
 *         your applications.  The code is provided 'AS IS', without warranty
 *         or condition of any kind.  IBM shall not be liable for any damages
 *         arising out of your use of the sample code, even if IBM has been
 *         advised of the possibility of such damages.
 */

#include "stdafx.h"     //Standard include added by Visual Studio
#include <memory.h>     //For the memset function
#include "CicsEci_2003.h"


#ifndef __cplusplus_cli

//Importing the EciWrapper namespace removes the need to prefix every member
//defined in the following code with EciWrapper::.
using namespace EciWrapper;


/*
 *      EciException class
 */
//ReturnCode property
EciReturnCode EciException::get_ReturnCode() {
    return this->rc;
}

//AbendCode property
String* EciException::get_AbendCode() {
    return this->abCode;
}

//The constructor calls the constructor of the base Exception class to set the
//exception message.
EciException::EciException(short retCode, String* abendCode) : Exception(S"The CICS ECI call failed.") {
    this->rc = (EciReturnCode) retCode;
    this->abCode = abendCode;
}


/*
 *      CicsSystem structure
 */
//Name property
String* CicsSystem::get_Name() {
    return this->sysName;
}

//Description property
String* CicsSystem::get_Description() {
    return this->sysDesc;
}

//Constructor
CicsSystem::CicsSystem(String* name, String* desc) {
    this->sysName = name;
    this->sysDesc = desc;
}


/*
 *      CicsEci class
 */
//CallType property - wraps eci_call_type
EciCallType CicsEci::get_CallType() {
    return (EciCallType) this->eciParms->eci_call_type;
}
void CicsEci::set_CallType(EciCallType callType) {
    this->eciParms->eci_call_type = (cics_sshort_t) callType;
}

//ProgramName property - wraps eci_program_name
String* CicsEci::get_ProgramName() {
    return this->NtvCharToMgdString(this->eciParms->eci_program_name, 8);
}
void CicsEci::set_ProgramName(String* programName) {
    this->MgdStringToNtvChar(programName, this->eciParms->eci_program_name, 8);
}

//UserID property - wraps eci_userid
String* CicsEci::get_UserID() {
    return this->NtvCharToMgdString(this->eciParms->eci_userid, 8);
}
void CicsEci::set_UserID(String* userID) {
    this->MgdStringToNtvChar(userID, this->eciParms->eci_userid, 8);
}

//Password property - wraps eci_password
String* CicsEci::get_Password() {
    return this->NtvCharToMgdString(this->eciParms->eci_password, 8);
}
void CicsEci::set_Password(String* password) {
    this->MgdStringToNtvChar(password, this->eciParms->eci_password, 8);
}

//TransactionID property - wraps eci_transid
String* CicsEci::get_TransactionID() {
    return this->NtvCharToMgdString(this->eciParms->eci_transid, 4);
}
void CicsEci::set_TransactionID(String* transID) {
    this->MgdStringToNtvChar(transID, this->eciParms->eci_transid, 4);
}

//AbendCode property - wraps eci_abend_code
String* CicsEci::get_AbendCode() {
    return this->NtvCharToMgdString(this->eciParms->eci_abend_code, 4);
}

//Commarea property - wraps eci_commarea
Byte CicsEci::get_Commarea() __gc[] {
    return this->commareaArray;
}
void CicsEci::set_Commarea(Byte commarea __gc[]) {
    this->commareaArray = commarea;
}

//Timeout property - wraps eci_timeout
short CicsEci::get_Timeout() {
    return this->eciParms->eci_timeout;
}
void CicsEci::set_Timeout(short timeout) {
    this->eciParms->eci_timeout = timeout;
}

//ExtendMode property - wraps eci_extend_mode
EciExtendMode CicsEci::get_ExtendMode() {
    return (EciExtendMode) this->eciParms->eci_extend_mode;
}
void CicsEci::set_ExtendMode(EciExtendMode extendMode) {
    this->eciParms->eci_extend_mode = (short) extendMode;
}

//Handle property - wraps eci_async_notify
IntPtr CicsEci::get_Handle() {
    return (IntPtr) this->eciParms->eci_async_notify.sem_handle;
}
void CicsEci::set_Handle(IntPtr handle) {
    this->eciParms->eci_async_notify.sem_handle = (cics_lhandle_t) handle.ToPointer();
}

//MessageID property - wraps eci_message_id
unsigned short CicsEci::get_MessageID() {
    return this->eciParms->eci_message_id;
}
void CicsEci::set_MessageID(unsigned short messageID) {
    this->eciParms->eci_message_id = messageID;
}

//MessageQualifier property - wraps eci_message_qualifier
short CicsEci::get_MessageQualifier() {
    return this->eciParms->eci_message_qualifier;
}
void CicsEci::set_MessageQualifier(short messageQualifier) {
    this->eciParms->eci_message_qualifier = messageQualifier;
}

//LUWToken property - wraps eci_luw_token
int CicsEci::get_LUWToken() {
    return this->eciParms->eci_luw_token;
}
void CicsEci::set_LUWToken(int luwToken) {
    this->eciParms->eci_luw_token = luwToken;
}

//SysID property - wraps eci_sysid
String* CicsEci::get_SysID() {
    return this->NtvCharToMgdString(this->eciParms->eci_sysid, 4);
}
void CicsEci::set_SysID(String* sysID) {
    this->MgdStringToNtvChar(sysID, this->eciParms->eci_sysid, 4);
}

//SystemName property - wraps eci_system_name
String* CicsEci::get_SystemName() {
    return this->NtvCharToMgdString(this->eciParms->eci_system_name, 8);
}
void CicsEci::set_SystemName(String* systemName) {
    this->MgdStringToNtvChar(systemName, this->eciParms->eci_system_name, 8);
}

//Callback property - wraps eci_callback
EciCallback* CicsEci::get_Callback() {
    return this->callbackDelegate;
}
void CicsEci::set_Callback(EciCallback* callback) {
    this->callbackDelegate = callback;
}

//TPN property - wraps eci_tpn
String* CicsEci::get_TPN() {
    return this->NtvCharToMgdString(this->eciParms->eci_tpn, 4);
}
void CicsEci::set_TPN(String* tpn) {
    this->MgdStringToNtvChar(tpn, this->eciParms->eci_tpn, 4);
}

//Constructor
CicsEci::CicsEci() {
    //Create the ECI_PARMS structure
    this->eciParms = new ECI_PARMS;
    memset(this->eciParms, 0, sizeof(ECI_PARMS));

    //Set ECI version
    this->eciParms->eci_version = ECI_VERSION_1A;

    //Set callback to null
    this->callbackDelegate = NULL;
}

//Destructors in Managed C++ compile to finalizers which are called when the
//garbage collector is about to dispose of the object.
CicsEci::~CicsEci() {
    //Delete ECI_PARMS
    delete this->eciParms;

    //Release delegate
    this->callbackDelegate = NULL;
}

//The MgdStringToNtvChar utility method copies the contents of a managed String
//object into a given char buffer, up to the size of the buffer.
void CicsEci::MgdStringToNtvChar(String* src, cics_char_t* dest, int len) {
    cics_char_t*   ntvChars;

    //Zero destination block
    memset(dest, 0, len);

    if (src != NULL) {
        //Get chars
        ntvChars = (cics_char_t*) Marshal::StringToHGlobalAnsi(src).ToPointer();

        //Copy to destination pointer
        memset(dest, 0, len);
        if (src->Length > len) {
            memcpy(dest, ntvChars, len);
        } else {
            memcpy(dest, ntvChars, src->Length);
        }

        //Free temp memory
        Marshal::FreeHGlobal((IntPtr) ntvChars);
    }
}

//The NtvCharToMgdString utility method creates a managed String object from a
//given unmanaged char buffer. The length of the text in the buffer is measured
//then a new String object is created and trailing whitespace is removed.
String* CicsEci::NtvCharToMgdString(cics_char_t* src, int len) {
    String* mgdString;

    //Measure the string up to a null character or the maximum length len
    int realLen = 0;

    while ((realLen < len) && (src[realLen] != 0)) {
        realLen++;
    }

    //Create managed string and trim off any trailing whitespace
    mgdString = __gc new String(src, 0, realLen);
    return mgdString->TrimEnd((S" ")->ToCharArray());
}

//The ListSystems method calls CICS_EciListSystems with an array large enough
//to hold 20 CICS systems. If the return code is ECI_ERR_MORE_SYSTEMS, the
//array size is increased to match the actual number of systems and the call
//is repeated. All error return codes except ECI_ERR_NO_SYSTEMS cause an
//EciException to be thrown. If there are no systems defined the method will
//continue as normal and an empty array will be returned.
CicsSystem CicsEci::ListSystems() __gc[] {
    CicsSystem          mgdSysArray __gc[];
    CICS_EciSystem_t*   ntvSysArray;
    unsigned short      servCount;
    short               ret;

    //Allocate space for 20 systems
    servCount = 20;
    ntvSysArray = new CICS_EciSystem_t[servCount];
    memset(ntvSysArray, 0, servCount * sizeof(CICS_EciSystem_t));

    //Make the list systems native call
    ret = CICS_EciListSystems(NULL, &servCount, ntvSysArray);

    //Test return code
    if (ret == ECI_ERR_MORE_SYSTEMS) {
        //More than 20 systems, so increase array size and try again
        delete [] ntvSysArray;
        ntvSysArray = new CICS_EciSystem_t[servCount];
        ret = CICS_EciListSystems(NULL, &servCount, ntvSysArray);
    }

    if ((ret != ECI_NO_ERROR) && (ret != ECI_ERR_NO_SYSTEMS)) {
        //An error occurred, throw an exception
        delete [] ntvSysArray;
        throw __gc new EciException(ret, String::Empty);
    }

    //Convert CICS_EciSystem_t array to CicsSystem array
    mgdSysArray = __gc new CicsSystem __gc[servCount];

    for (int i = 0; i < servCount; i++) {
        mgdSysArray[i] = CicsSystem(
                        CicsEci::NtvCharToMgdString(ntvSysArray[i].SystemName, CICS_ECI_SYSTEM_MAX),
                        CicsEci::NtvCharToMgdString(ntvSysArray[i].Description, CICS_ECI_DESCRIPTION_MAX)
                      );
    }

    //Delete native array
    delete [] ntvSysArray;

    //Return managed array
    return mgdSysArray;
}

//The Execute method prepares eci_commarea, eci_commarea_length and
//eci_callback members of the ECI_PARMS structure before making the
//CICS_ExternalCall function call. An additional reference to the commarea
//memory is kept in the form of a pinned pointer, which prevents the garbage
//collector from moving or freeing the memory until the pointer goes out of
//scope at the end of the method. This allows unmanaged code to access the
//data referenced by the pointer safely.
void CicsEci::Execute() {
    short       ret;
    Byte __pin* commareaPtr;

    //Pin the managed commarea and assign to ECI_PARMS
    if (this->commareaArray != NULL) {
        commareaPtr = &this->commareaArray[0];
        this->eciParms->eci_commarea = commareaPtr;
        this->eciParms->eci_commarea_length = this->commareaArray->Length;
    }

    //Set up the callback if needed
    if (this->callbackDelegate != NULL) {
        this->eciParms->eci_callback = (CICS_EciNotify_t) Marshal::GetFunctionPointerForDelegate(this->callbackDelegate).ToPointer();
    }

    //Make the CICS call
    ret = CICS_ExternalCall(this->eciParms);

    //Check the return code
    if (ret != ECI_NO_ERROR) {
        //Throw an exception
        Exception*  ex;
        ex = __gc new Exception(String::Format(S"Return code was {0}", __box(ret)));
        throw ex;
    }
}

#endif